@@ -4,8 +4,9 @@
 import boto3
 import logging
 import time
-import pandas as pd
 import os
+from .task_queue import TaskQueue
+from dativa.tools.pandas import CSVHandler
 
 logger = logging.getLogger("dativa.tools.aws")
 
@@ -24,13 +25,13 @@ class AthenaClientError(Exception):
         self.reason = reason
 
 
-class AthenaClient():
+class AthenaClient(TaskQueue):
     """
     A client for AWS Athena that will create tables from S3 buckets (using AWS Glue)
     and run queries against these tables.
     """
 
-    def __init__(self, region, db, max_queries=3, max_retries=3):
+    def __init__(self, region, db, max_queries=3, max_retries=3, s3_parquet=None):
         """
         Create an AthenaClient
 
@@ -43,104 +44,53 @@ class AthenaClient():
 
         """
         self.athena = boto3.client(service_name='athena', region_name=region)
+        self.glue = boto3.client(service_name='glue', region_name=region,
+                                 endpoint_url='https://glue.{0}.amazonaws.com'.format(region))
         self.db_name = db
-        self.max_queries = max_queries
-        self.queue = []
         self.aws_region = region
-        self.max_retries = max_retries
+        self.scp = s3_parquet
 
-    @property
-    def is_active(self):
-        """
-        Returns True if there are active or pending queries in the queue.
-        Processes the queue before returning a result
-        """
-        self._process_queue()
-        return self.number_active + self.number_pending > 0
+        super(AthenaClient, self).__init__(max_queries, max_retries)
 
-    @property
-    def number_active(self):
-        """
-        Returns the number of active queries being processed by Athena
-        """
-        i = 0
-        for query in self.queue:
-            if query.is_active:
-                i = i + 1
-        return i
-
-    @property
-    def number_pending(self):
-        """
-        Returns the number of pending queries that have not yet been
-        sent to Athena
-        """
-        i = 0
-        for query in self.queue:
-            if query.is_pending:
-                i = i + 1
-        return i
-
-    def _get_query_status(self, query):
+    def _update_task_status(self, task):
         """
         Gets the status of the query, and updates its status in the queue.
         Any queries that fail are reset to pending so they will be run a second time
         """
 
-        logger.info("...checking status of query {0}".format(query.name))
-
-        status = self.athena.get_query_execution(QueryExecutionId=query.id)["QueryExecution"]["Status"]
+        logger.debug("...checking status of query {0}".format(task.name))
+        status = self.athena.get_query_execution(QueryExecutionId=task.id)["QueryExecution"]["Status"]
 
         if status["State"] == "RUNNING":
-            query.is_active = True
+            task.is_complete = False
         elif status["State"] == "SUCCEEDED":
-            query.is_active = False
+            task.is_complete = True
+            if task.arguments["parquet"]:
+                logger.info("starting conversion to")
+                self.scp.convert("{0}{1}.csv".format(task.arguments["output_location"],
+                                                     task.id),
+                                 delete_csv=True,
+                                 name="convert {0}".format(task.name))
         else:
-            if query.retries < self.max_retries:
-                logger.info("Retrying query {0} completed with state {1}".format(query.name, status["State"]))
-                if "StateChangeReason" in status:
-                    logger.info(status["StateChangeReason"])
-
-                query.is_active = False
-                query.is_pending = True
-                query.retries += 1
+            if "StateChangeReason" in status:
+                task.error = status["StateChangeReason"]
             else:
-                logger.info("Query {0} exceeded maximum retry limit: Marking as failed".format(query.name))
-                query.is_active = False
+                task.error = status["State"]
 
-
-        return query
-
-    def _query_athena(self, query):
+    def _trigger_task(self, task):
         """
         Runs a query in Athena
         """
 
-        logger.info("Starting query {0}".format(query.name))
+        logger.info("Starting query {0} to {1}".format(task.name, task.arguments["output_location"]))
 
-        query.id = self.athena.start_query_execution(
-            QueryString=query.sql,
+        task.id = self.athena.start_query_execution(
+            QueryString=task.arguments["sql"],
             QueryExecutionContext={'Database': self.db_name},
-            ResultConfiguration={'OutputLocation': query.output_location}
+            ResultConfiguration={'OutputLocation': task.arguments["output_location"]}
         )["QueryExecutionId"]
 
-        query.is_active = True
-        query.is_pending = False
-        return query
-
-    def _process_queue(self):
-        """
-        Checks the status of each active query in the queue and starts
-        any news ones that are pending.
-        """
-        for ix, query in enumerate(self.queue):
-            if query.is_active:
-                query = self._get_query_status(query)
-                self.queue[ix] = query
-            elif query.is_pending and self.number_active < self.max_queries:
-                    self.queue[ix] = self._query_athena(query)
-
-    def add_query(self, sql, name, output_location):
+    def add_query(self, sql, name, output_location, parquet=False):
         """
         Adds a query to Athena. Respects the maximum number of queries specified when the module was created.
         Retries queries when they fail so only use when you are sure your syntax is correct!
@@ -153,35 +103,38 @@ class AthenaClient():
         - location - the S3 prefix where you want the results stored
 
         """
-        query = Object()
-        query.sql = sql
-        query.name = name
-        query.output_location = output_location
-        query.is_active = False
-        query.is_pending = True
-        query.id = None
-        query.retries = 0
+        if parquet is True and self.scp is None:
+            raise AthenaClientError("Cannot output in Parquet without a S3Csv2Parquet object")
+
+        query = self.add_task(name=name,
+                              args={"sql": sql,
+                                    "output_location": output_location,
+                                    "parquet": parquet})
 
-        self.queue.append(query)
-        if self.number_active < self.max_queries:
-            self._process_queue()
         return query
 
     def wait_for_completion(self):
+        """Wait cor completion of any queries and also any pending parquet conversions
+        :rtype: object
         """
-        Waits for any pending queries to be completed
-        """
-        while self.is_active:
-            time.sleep(10)
+        super(AthenaClient, self).wait_for_completion()
+        if self.scp is not None:
+            self.scp.wait_for_completion()
+
+    def _db_exists(self):
+        for database in self.glue.get_databases(MaxResults=1000)["DatabaseList"]:
+            if database["Name"] == self.db_name:
+                return True
+        return False
 
     def create_table(self,
                      crawler_target,
                      table_name,
                      columns=None,
-                     schema_change_policy={'UpdateBehavior': 'UPDATE_IN_DATABASE', 'DeleteBehavior': 'DEPRECATE_IN_DATABASE'},
+                     schema_change_policy=None,
                      aws_role='AWSGlueServiceRoleDefault',
                      serde=None,
-                     serde_parameters=[]):
+                     serde_parameters=None):
         """
         Creates a table in AWS Glue using a crawler.
 
@@ -206,47 +159,53 @@ class AthenaClient():
 
         """
 
-        crawler_name = "{0}_{1}_crawler".format(self.db_name, table_name)
+        if schema_change_policy is None:
+            schema_change_policy = {'UpdateBehavior': 'UPDATE_IN_DATABASE', 'DeleteBehavior': 'DEPRECATE_IN_DATABASE'}
 
-        # create a glue client...
-        glue = boto3.client(service_name='glue', region_name=self.aws_region,
-                            endpoint_url='https://glue.{0}.amazonaws.com'.format(self.aws_region))
+        if serde_parameters is None:
+            serde_parameters = []
+
+        crawler_name = "{1}_{0}_crawler".format(self.db_name, table_name)
+
+        # creare the database if it exists
+        if not self._db_exists():
+            self.glue.create_database(DatabaseInput={"Name": self.db_name})
 
         # if the crawler exists then update it:
         have_updated = False
-        for crawler in glue.get_crawlers(MaxResults=1000)["Crawlers"]:
+        for crawler in self.glue.get_crawlers(MaxResults=1000)["Crawlers"]:
             if crawler_name == crawler["Name"]:
                 logger.info("updating crawler {0}".format(crawler_name))
-                glue.update_crawler(Name=crawler_name,
-                                    Role=aws_role,
-                                    Targets=crawler_target,
-                                    DatabaseName=self.db_name,
-                                    Classifiers=[],
-                                    SchemaChangePolicy=schema_change_policy)
+                self.glue.update_crawler(Name=crawler_name,
+                                         Role=aws_role,
+                                         Targets=crawler_target,
+                                         DatabaseName=self.db_name,
+                                         Classifiers=[],
+                                         SchemaChangePolicy=schema_change_policy)
                 have_updated = True
                 break
 
         # otherwise create it
         if not have_updated:
             logger.info("creating crawler {0} ".format(crawler_name))
-            glue.create_crawler(Name=crawler_name,
-                                Role=aws_role,
-                                Targets=crawler_target,
-                                DatabaseName=self.db_name,
-                                Classifiers=[],
-                                SchemaChangePolicy=schema_change_policy)
-
+            self.glue.create_crawler(Name=crawler_name,
+                                     Role=aws_role,
+                                     Targets=crawler_target,
+                                     DatabaseName=self.db_name,
+                                     Classifiers=[],
+                                     SchemaChangePolicy=schema_change_policy)
+ 
         # start the crawler and wait for it to complete:
         logger.info("starting crawler {0}".format(crawler_name))
-        glue.start_crawler(Name=crawler_name)
-        while glue.get_crawler(Name=crawler_name)["Crawler"]["State"] != "READY":
+        self.glue.start_crawler(Name=crawler_name)
+        while self.glue.get_crawler(Name=crawler_name)["Crawler"]["State"] != "READY":
             logger.info("... waiting for crawler {0} to finish".format(crawler_name))
             time.sleep(5)
 
         if columns is not None or serde is not None:
             # get the table and update the column names
             logger.info("updating tables {0}".format(table_name))
-            table = glue.get_table(DatabaseName=self.db_name, Name=table_name)["Table"]
+            table = self.glue.get_table(DatabaseName=self.db_name, Name=table_name)["Table"]
             
             if columns is not None:
                 table["StorageDescriptor"]["Columns"] = columns
@@ -260,15 +219,15 @@ class AthenaClient():
             else:
                 table["StorageDescriptor"]["SerdeInfo"]["Parameters"] = serde_parameters
 
-            glue.update_table(DatabaseName=self.db_name,
-                              TableInput={'Name': table_name,
-                                          'StorageDescriptor': table["StorageDescriptor"],
-                                          'PartitionKeys': table["PartitionKeys"],
-                                          'TableType': table["TableType"],
-                                          'Parameters': table["Parameters"]})
+            self.glue.update_table(DatabaseName=self.db_name,
+                                   TableInput={'Name': table_name,
+                                               'StorageDescriptor': table["StorageDescriptor"],
+                                               'PartitionKeys': table["PartitionKeys"],
+                                               'TableType': table["TableType"],
+                                               'Parameters': table["Parameters"]})
 
             # now check for partitions
-            partitions = glue.get_partitions(DatabaseName=self.db_name, TableName=table_name)
+            partitions = self.glue.get_partitions(DatabaseName=self.db_name, TableName=table_name)
 
             for partition in partitions["Partitions"]:
                 logger.info("Updating partition {0}".format(partition["Values"]))
@@ -284,22 +243,28 @@ class AthenaClient():
                 else:
                     partition["StorageDescriptor"]["SerdeInfo"]["Parameters"] = serde_parameters
 
-                glue.update_partition(DatabaseName=self.db_name,
-                                      TableName=table_name,
-                                      PartitionValueList=partition["Values"],
-                                      PartitionInput={'StorageDescriptor': partition["StorageDescriptor"],
-                                                      'Values': partition["Values"]})
+                self.glue.update_partition(DatabaseName=self.db_name,
+                                           TableName=table_name,
+                                           PartitionValueList=partition["Values"],
+                                           PartitionInput={'StorageDescriptor': partition["StorageDescriptor"],
+                                                           'Values': partition["Values"]})
+        # delete the crawler...
+        self.glue.delete_crawler(Name=crawler_name)
 
     def get_query_result(self, query):
         """
-        Returns Pandas dataframe containing query result if query has completed
+        Returns Pandas DataFrame containing query result if query has completed
         """
-        query = self._get_query_status(query)
+        self._update_task_status(query)
 
-        if query.is_active is False and query.is_pending is False:
-            filepath = os.path.join(query.output_location, "{}.csv".format(query.id))
-            logger.info("Fetching results from {}".format(filepath))
-            df = pd.read_csv(filepath)
-            return df
+        if query.is_complete:
+            if query.error:
+                raise AthenaClientError("Cannot fetch results since query failed")
+            else:
+                filepath = os.path.join(query.arguments["output_location"], "{}.csv".format(query.id))
+                logger.info("Fetching results from {}".format(filepath))
+                csv = CSVHandler()
+                df = csv.load_df(filepath)
+                return df
         else:
             raise AthenaClientError("Cannot fetch results since query hasn't completed")
