@@ -1,6 +1,7 @@
 from .date_time import is_numeric, string_to_datetime, datetime_to_string, format_string_is_valid
 from .csv import CSVHandler, FpCSVEncodingError
 from .columns import get_column_name, get_unique_column_name
+from .parquet_handler import ParquetHandler
 
 __all__ = ['is_numeric',
            'string_to_datetime',
@@ -9,5 +10,6 @@ __all__ = ['is_numeric',
            'CSVHandler',
            'FpCSVEncodingError',
            'get_column_name',
-           'get_unique_column_name'
-           ]
+           'get_unique_column_name',
+           'ParquetHandler'
+          ]
