@@ -5,10 +5,12 @@
 import ast
 import pandas as pd
 import logging
-from io import StringIO, BytesIO
+from io import StringIO, BytesIO, TextIOWrapper
 from csv import Sniffer, Error
 from chardet.universaldetector import UniversalDetector
-
+import boto3
+from urllib.parse import urlparse
+import s3fs
 
 logger = logging.getLogger("dativa.tools.pandas.csv")
 
@@ -42,6 +44,7 @@ class CSVHandler():
     DEFAULT_DELIMITER = ","
     DEFAULT_HEADER = 0
     DEFAULT_QUOTECHAR = "\""
+    S3_PREFIX = "s3://"
 
     def __init__(self,
                  detect_parameters=False,
@@ -79,8 +82,10 @@ class CSVHandler():
         else:
             header = None
 
+        full_path = self.base_path + file
+
         if force_dtype is not None:
-            df = pd.read_csv(self.base_path + file,
+            df = pd.read_csv(full_path,
                              encoding=self.csv_encoding,
                              sep=ast.literal_eval("'%s'" % self.csv_delimiter),
                              quotechar=ast.literal_eval(
@@ -90,7 +95,7 @@ class CSVHandler():
                              skip_blank_lines=False,
                              dtype=force_dtype)
         else:
-            df = pd.read_csv(self.base_path + file,
+            df = pd.read_csv(full_path,
                              encoding=self.csv_encoding,
                              sep=ast.literal_eval("'%s'" % self.csv_delimiter),
                              quotechar=ast.literal_eval(
@@ -118,9 +123,19 @@ class CSVHandler():
                 self.csv_quotechar == self.DEFAULT_QUOTECHAR):
             logger.debug("sniffing file type")
 
-            # create a sample and get the encoding...
-            with open(self.base_path + file, mode="rb") as f:
-                sample = f.read(1024 * 1024)
+            # create a sample...
+            full_path = self.base_path + file
+            if full_path.startswith(self.S3_PREFIX):
+                # This is an S3 location
+                logger.debug("s3 location")
+                fs = s3fs.S3FileSystem(anon=False)
+                with fs.open(full_path, "rb") as f:
+                    sample = f.read(1024 * 1024)
+            else:
+                with open(full_path, mode="rb") as f:
+                    sample = f.read(1024 * 1024)
+
+            # get the encoding...
             self.csv_encoding = self._get_encoding(sample)
             if self.csv_encoding is None:
                 self.csv_encoding = 'windows-1252'
@@ -167,11 +182,30 @@ class CSVHandler():
         except (UnicodeDecodeError, pd.errors.ParserError) as e:
             if self.auto_detect:
                 if self._sniff_parameters(file):
-                    self.save()
                     logger.debug("second attempt to load")
                     return self._get_df_from_raw(file, force_dtype)
             raise FpCSVEncodingError(e)
 
+    def _s3_path_to_bucket(self, url):
+
+        s3_bucket = None
+        s3_key = ""
+
+        if url:
+            parsed_s3_url = urlparse(url)
+            if parsed_s3_url and parsed_s3_url.scheme.lower() == "s3":
+                s3_bucket = parsed_s3_url.netloc
+                if parsed_s3_url.path:
+                    s3_key = parsed_s3_url.path.lstrip('/')
+
+        return s3_bucket, s3_key
+
+    def load_df(self, file, force_dtype=None):
+        """
+        Synonym for 'get_dataframe' for consistency with 'save_df'
+        """
+        return self.get_dataframe(file, force_dtype)
+
     def get_dataframe(self, file, force_dtype=None):
         """
         Opens a CSV file using the specified configuration for the class
@@ -185,27 +219,63 @@ class CSVHandler():
     def save_df(self, df, file):
         """
         Writes a formatted string from a dataframe using the specified
-        configuration for the class the file
+        configuration for the class the file. Detects if base_path is
+        an S3 location and saves data there if required.
         """
 
         if self.csv_header == -1:
             header = False
         else:
             header = True
-        file = df.to_csv(file,
-                         index=False,
-                         encoding=self.csv_encoding,
-                         sep=ast.literal_eval("'%s'" % self.csv_delimiter),
-                         quotechar=ast.literal_eval(
-                             "'%s'" % self.csv_quotechar),
-                         header=header)
+
+        full_path = self.base_path + file
+
+        if full_path.startswith(self.S3_PREFIX):
+            # This is an S3 location
+            s3 = boto3.client('s3')
+            buffer = StringIO()
+            # Stream the data via buffer
+            df.to_csv(buffer,
+                      index=False,
+                      encoding=self.csv_encoding,
+                      sep=ast.literal_eval("'%s'" % self.csv_delimiter),
+                      quotechar=ast.literal_eval(
+                            "'%s'" % self.csv_quotechar),
+                      header=header)
+            buffer.seek(0)
+            # Extract bucket name from base path
+            s3_bucket, s3_key = self._s3_path_to_bucket(full_path)
+            # Save buffered data to file in S3 bucket
+            if s3_bucket:
+                s3.put_object(Bucket=s3_bucket,
+                              Key=s3_key,
+                              ContentEncoding=self.csv_encoding,
+                              ContentType="text/csv",
+                              Body=buffer.read().encode(self.csv_encoding))
+        else:
+            df.to_csv(full_path,
+                      index=False,
+                      encoding=self.csv_encoding,
+                      sep=ast.literal_eval("'%s'" % self.csv_delimiter),
+                      quotechar=ast.literal_eval("'%s'" % self.csv_quotechar),
+                      header=header)
 
     def df_to_string(self, df):
         """
         Returns a formatted string from a dataframe using the specified
         configuration for the class
         """
+        
+        if self.csv_header == -1:
+            header = False
+        else:
+            header = True
 
         buffer = StringIO()
-        self.save_df(df, buffer)
+        df.to_csv(buffer,
+                  index=False,
+                  encoding=self.csv_encoding,
+                  sep=ast.literal_eval("'%s'" % self.csv_delimiter),
+                  quotechar=ast.literal_eval("'%s'" % self.csv_quotechar),
+                  header=header)
         return buffer.getvalue()
